from operator import attrgetter

from pokercore.exceptions import HandCreationError, HandComparisonError
from pokercore.card import Card, CardCreationError


class Hand(object):
    """Class representing a poker hand.

    A poker hand consists of one or more Card objects, passed to the
    constructor contained in some iterable. Its main attributes are two:

    * value - an integer between 0 and 8 representing the category of
              the poker hand
    * best_cards - the best (at most 5) cards that consist the actual hand

    A Hand can be compared to other Hand objects, judging by the value,
    and then the best cards, lexicographically.
    """

    names = ('high card',
             'one pair',
             'two pair',
             'three of a kind',
             'straight',
             'flush',
             'full house',
             'four of a kind',
             'straight flush')

    def __init__(self, cards):
        self.cards = list(cards)
        if not self.cards:
            raise HandCreationError('cannot create empty hand')

        for c in self.cards:
            if not isinstance(c, Card):
                raise HandCreationError('hand can only consist of cards')

        self.value, self.best_cards = self._evaluate()

    def __repr__(self):
        return 'Hand(%s: %s)' % (self.names[self.value],
                                 ', '.join(str(c) for c in self.best_cards))

    def _find_straight(self, cards):
        """find highest possible straight in given uniquely ranked cards"""

        connections = 0
        current = cards[0]
        for i in xrange(1, len(cards)):
            if current == cards[i] + 1:
                connections += 1
                if connections == 4:
                    return cards[i - 4:i + 1]
            else:
                connections = 0
            current = cards[i]

        possible = cards[-4:] + [cards[0]]
        if possible == [3, 2, 1, 0, 12]:
            return possible
        return []

    def _max_group(self, length, *excluded, **kwargs):
        """find highest-ranked group of equally ranked cards,
        with minimum length as the one given
        """

        result = []
        for g in self._groups:
            if len(g) >= length and g[0] not in excluded and g > result:
                result = g[:5]

        if kwargs.get('one'):
            result = result[:1]

        return result

    def _kickers(self, n, *excluded):
        """find at most n unequally ranked kickers"""

        result = []
        for i in xrange(n):
            result.extend(self._max_group(1,
                                          *(result + list(excluded)),
                                          one=True))
        return result

    def _evaluate(self):
        """evaluate the kind of the hand and find the best cards"""

        # straight flush (checking also for flush)
        cards = sorted(self.cards, key=attrgetter('suit', 'rank'),
                       reverse=True)
        by_suit = []
        suit = -1
        for c in cards:
            if c.suit != suit:
                by_suit.append([])
                suit = c.suit
            by_suit[-1].append(c)

        flush = []
        straight_flush = []
        for suited in by_suit:
            if len(suited) >= 5:
                if suited > flush:
                    flush = suited[:5]
                straight_flush = max(self._find_straight(suited),
                                     straight_flush)

        if straight_flush:
            return 8, straight_flush

        # group equally ranked cards
        groups = {}
        for c in self.cards:
            groups.setdefault(c.rank, []).append(c)
        self._groups = groups.values()

        # four of a kind
        four_of_a_kind = self._max_group(4)
        if four_of_a_kind:
            return 7, four_of_a_kind + self._kickers(1, four_of_a_kind[0])

        # full house (checking also for three of a kind)
        three_of_a_kind = self._max_group(3)
        if three_of_a_kind:
            pair = self._max_group(2, three_of_a_kind[0])
            if pair:
                return 6, three_of_a_kind + pair

        # flush
        if flush:
            return 5, flush

        # straight
        uniquely_ranked = []
        unique_ranks = set()
        for c in sorted(self.cards, reverse=True):
            if c.rank not in unique_ranks:
                uniquely_ranked.append(c)
                unique_ranks.add(c.rank)

        straight = self._find_straight(uniquely_ranked)
        if straight:
            return 4, straight

        # three of a kind
        if three_of_a_kind:
            return 3, three_of_a_kind + self._kickers(2, three_of_a_kind[0])

        # pairs
        pair = self._max_group(2)
        if pair:
            second_pair = self._max_group(2, pair[0])
            if second_pair:
                # two pair
                return 2, (pair + second_pair +
                           self._kickers(1, pair[0], second_pair[0]))
            # one pair
            return 1, pair + self._kickers(3, pair[0])

        # high card
        return 0, uniquely_ranked[:5]

    def __cmp__(self, other):
        if not isinstance(other, Hand):
            raise HandComparisonError('cannot compare hand to non-hand')

        return cmp((self.value, self.best_cards),
                   (other.value, other.best_cards))

    @classmethod
    def from_chars(cls, *args):
        """return a new object from pairs of character symbols

        works with either multiple arguments, or a single iterable
        """

        try:
            return cls(Card.from_chars(chars) for chars in args)
        except CardCreationError:
            try:
                return cls(Card.from_chars(chars) for chars in args[0])
            except (TypeError, CardCreationError):
                pass
            raise HandCreationError('invalid card symbols')


from os.path import dirname, join
from unittest import TestCase

from pokercore.card import Card
from pokercore.hand import Hand, HandCreationError, HandComparisonError


class TestHand(TestCase):

    def setUp(self):
        self.players = [], []
        self.wins = []
        with open(join(dirname(__file__), 'testcases')) as testcases:
            for i, line in enumerate(testcases):
                case = i % 3
                parts = line.split()
                if case == 2:
                    self.wins.append(int(parts[0]))
                else:
                    self.players[case].append((Hand.from_chars(parts[:-1]),
                                               int(parts[-1])))

    def test_create(self):
        self.assertRaises(TypeError, Hand, Card(1, 2))
        self.assertRaises(HandCreationError, Hand, [])
        self.assertRaises(HandCreationError, Hand, [3, 1])

        self.assertRaises(HandCreationError, Hand.from_chars)
        self.assertRaises(HandCreationError, Hand.from_chars, 4)
        self.assertRaises(HandCreationError, Hand.from_chars, 'aS', '4p')
        self.assertRaises(HandCreationError, Hand.from_chars, ['Th', 9])
        self.assertRaises(HandCreationError, Hand.from_chars,
                          'KH', 'Qs', 'Acd')

        hand = Hand([Card(4, 2), Card.from_chars('AS'), Card(9, 0)])
        self.assertIsInstance(hand, Hand)
        self.assertEqual(hand.value, 0)
        self.assertEqual(Hand.names[hand.value], 'high card')
        self.assertEqual(hand.best_cards, [12, 9, 4])

        hand = Hand.from_chars('4s', 'Th', '3s', '3c', 'Js', '3d')
        self.assertIsInstance(hand, Hand)
        self.assertEqual(hand.value, 3)
        self.assertEqual(Hand.names[hand.value], 'three of a kind')
        self.assertEqual(hand.best_cards, [1, 1, 1, 9, 8])

        hand = Hand.from_chars(['KH', '9H', 'KS', '9C'])
        self.assertIsInstance(hand, Hand)
        self.assertEqual(hand.value, 2)
        self.assertEqual(Hand.names[hand.value], 'two pair')
        self.assertEqual(hand.best_cards, [11, 11, 7, 7])

    def test_value(self):
        for i in xrange(0, 2):
            for h in self.players[i]:
                self.assertEqual(h[0].value, h[1])

    def test_compare(self):
        hand = Hand([Card.from_chars('Qc'), Card(1, 3)])
        self.assertRaises(HandComparisonError, lambda h: h < Card(5, 2), hand)
        self.assertRaises(HandComparisonError, lambda h: h >= 0, hand)
        self.assertRaises(HandComparisonError, lambda h: h != 'Qc', hand)
        self.assertRaises(HandComparisonError, lambda h: h == 10, hand)
        self.assertEqual(hand, hand)

        for i in xrange(len(self.players[0])):
            first = self.players[0][i][0]
            second = self.players[1][i][0]
            self.assertEqual(cmp(first, second), self.wins[i])


from unittest import TestCase

from pokercore.card import Card
from pokercore.deck import Deck, DeckNotIntegerError, DeckTooManyError


class TestDeck(TestCase):

    def test_create(self):
        deck = Deck()
        self.assertEqual(len(deck.cards), 52)
        self.assertEqual(len(deck.remaining), 52)
        for i in xrange(52):
            self.assertIsInstance(deck.cards[i], Card)
            self.assertIsInstance(deck.remaining[i], Card)
        self.assertEqual(sorted(deck.cards), sorted(set(deck.cards)))
        self.assertEqual(sorted(deck.remaining), sorted(set(deck.remaining)))

    def test_draw(self):
        deck = Deck()
        self.assertRaises(DeckNotIntegerError, deck.draw, '7')
        self.assertRaises(DeckTooManyError, deck.draw, 53)

        card = deck.draw()
        self.assertIsInstance(card, Card)
        self.assertEqual(len(deck.remaining), 51)

        cards = deck.draw(4)
        for c in cards:
            self.assertIsInstance(c, Card)
            for r in deck.remaining:
                self.assertFalse(c.identical_to(r))
        self.assertEqual(len(cards), 4)
        self.assertEqual(len(deck.remaining), 47)

    def test_shuffle(self):
        deck = Deck()
        deck.draw(10)
        deck.shuffle()
        self.assertEqual(len(deck.remaining), 52)


"""pokercore

A poker engine core, in Python

pokercore provides 3 classes to be used in a poker engine.
Card (a playing card), Hand (a poker hand consisting of Cards,
with evaluation capabilities) and Deck (a deck of Cards).

Its is a simple starter, mainly written for exploring purposes,
but can be extended and/or used to build something bigger. It is
released under the MIT license.
"""

__title__ = 'pokercore'
__version__ = '0.1.2'
__author__ = 'Aristotelis Mikropoulos'

from pokercore.exceptions import *
from pokercore.card import Card
from pokercore.hand import Hand
from pokercore.deck import Deck



from unittest import TestCase

from pokercore.card import (Card,
                            CardCreationError,
                            CardArithmeticError,
                            CardComparisonError)


class TestCard(TestCase):

    def test_create(self):
        self.assertRaises(CardCreationError, Card, '3', 2)
        self.assertRaises(CardCreationError, Card, 9, 'H')
        self.assertRaises(CardCreationError, Card, 3, 5)
        self.assertRaises(CardCreationError, Card, 13, 0)
        self.assertRaises(CardCreationError, Card.from_chars, '51')
        self.assertRaises(CardCreationError, Card.from_chars, '10S')

        card = Card(3, 2)
        self.assertIsInstance(card, Card)
        self.assertEqual(card.rank, 3)
        self.assertEqual(card.suit, 2)

        card = Card.from_chars('TC')
        self.assertIsInstance(card, Card)
        self.assertEqual(card.rank, 8)
        self.assertEqual(card.suit, 0)

    def test_arithmetic(self):
        first = Card.from_chars('8D')
        self.assertRaises(TypeError, lambda c: c * 4, first)
        self.assertRaises(TypeError, lambda c: 1 * c, first)
        self.assertRaises(TypeError, lambda c: c / 2, first)
        self.assertRaises(TypeError, lambda c: 8 / c, first)
        self.assertRaises(CardArithmeticError, lambda c: c + '3', first)
        self.assertRaises(CardArithmeticError, lambda c: c - '1', first)

        self.assertEqual(first + 3, 9)
        self.assertEqual(2 + first, 8)
        self.assertEqual(first - 4, 2)
        self.assertEqual(10 - first, 4)

        second = Card(2, 2)
        self.assertRaises(TypeError, lambda c: c * second, first)
        self.assertRaises(TypeError, lambda c: second / c, first)
        self.assertEqual(first + second, 8)
        self.assertEqual(first - second, 4)
        self.assertEqual(second + 3 - first, -1)
        self.assertEqual(10 - second + first, 14)

    def test_compare(self):
        first = Card(9, 2)
        self.assertRaises(CardComparisonError, lambda c: c > 'A', first)
        self.assertRaises(CardComparisonError, lambda c: 't' <= c, first)
        self.assertRaises(CardComparisonError, lambda c: c == '9', first)
        self.assertRaises(CardComparisonError, lambda c: c != 'j', first)
        self.assertRaises(CardComparisonError, first.identical_to, 9)

        second = Card(11, 1)
        self.assertNotEqual(first, second)
        self.assertLess(first, second)
        self.assertLessEqual(first, second)
        self.assertGreater(second, first)
        self.assertGreaterEqual(second, first)
        self.assertFalse(first.identical_to(second))

        second = Card.from_chars('js')
        self.assertEqual(first, second)
        self.assertFalse(second.identical_to(first))

        second = Card.from_chars('Jh')
        self.assertEqual(second, first)
        self.assertTrue(first.identical_to(second))

        self.assertEqual(first, 9)
        self.assertLess(first, 11)
        self.assertLessEqual(first, 12)
        self.assertGreater(first, 2)
        self.assertGreaterEqual(first, 9)


import random

from pokercore.exceptions import DeckNotIntegerError, DeckTooManyError
from pokercore.card import Card


class Deck(object):
    """Class representing a deck of cards.

    A Deck consists of 52 Card objects, starting shuffled.
    It has two methods:

    * shuffle - Restore all cards to the deck, then shuffle it.
    * draw - Draw n cards from the top of the deck, returning them
             as a list of Card objects. If n is not provided, draw
             and return one Card object.
    """

    def __init__(self):
        self.cards = [Card(r, s) for r in xrange(13) for s in xrange(4)]
        self.shuffle()

    def shuffle(self):
        """restore all cards to deck, then shuffle it"""

        self.remaining = list(self.cards)
        random.shuffle(self.remaining)

    def draw(self, n=None):
        """remove and return a card from the deck,
        or a list of n cards, if n is given
        """

        if n is None:
            return self.remaining.pop()

        if not isinstance(n, int):
            raise DeckNotIntegerError('draw count must be an integer')

        if n > len(self.remaining):
            raise DeckTooManyError('cannot draw more than remaining')

        return [self.remaining.pop() for i in xrange(n)]


class PokerError(Exception):
    """Generic poker error"""


class CardError(PokerError):
    """Card-related error"""


class HandError(PokerError):
    """Hand-related error"""


class DeckError(PokerError):
    """Deck-related error"""


class CardCreationError(CardError):
    """Attempt to create a Card object using invalid arguments"""


class CardArithmeticError(CardError):
    """Invalid arithmetic operation on a Card object"""


class CardComparisonError(CardError):
    """Invalid comparison of a Card object"""


class HandCreationError(HandError):
    """Attempt to create a Hand object using invalid arguments"""


class HandComparisonError(HandError):
    """Invalid comparison of a Hand object"""


class DeckNotIntegerError(DeckError):
    """Given draw count is not an integer"""


class DeckTooManyError(DeckError):
    """Given draw count is greater than the count of remaining cards"""


from setuptools import setup

with open('README') as readme:
    long_description = readme.read()

setup(
    name='pokercore',
    version='0.1.2',
    description='A poker engine core, in Python',
    long_description=long_description,
    author='Aristotelis Mikropoulos',
    author_email='amikrop@gmail.com',
    url='https://pypi.python.org/pypi/pokercore',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Games/Entertainment',
    ],
    packages=['pokercore', 'pokercore.test'],
    package_data={'pokercore.test': ['testcases']},
)



from pokercore.exceptions import (CardCreationError,
                                  CardArithmeticError, CardComparisonError)


class Card(object):
    """Class representing a playing card.

    A playing card consists of two integers, passed to the constructor during
    instantiation:

    * rank - the rank of the card [0-12]
    * suit - the suit of the card [0-3]

    A Card can be compared to, added to, subtracted to and subtracted by
    other Card objects and integers (using their ranks, resulting in plain
    integers). For identity check, the identical_to method is provided.
    """

    ranks = ('2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K', 'A')
    suits = ('c', 'd', 'h', 's')

    def __init__(self, rank, suit):
        if rank not in xrange(13):
            raise CardCreationError('card rank must lie between'
                                    '0 and 12, inclusive')

        if suit not in xrange(4):
            raise CardCreationError('card suit must lie between'
                                    '0 and 3, inclusive')

        self.rank = rank
        self.suit = suit

    def __repr__(self):
        return 'Card(%s%s)' % (self.ranks[self.rank], self.suits[self.suit])

    def __add__(self, other):
        if isinstance(other, int):
            return self.rank + other

        if isinstance(other, Card):
            return self.rank + other.rank

        raise CardArithmeticError('cards can only be added'
                                  'to cards and integers')

    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, int):
            return self.rank - other

        if isinstance(other, Card):
            return self.rank - other.rank

        raise CardArithmeticError('cards can only be subtracted'
                                  'by cards and integers')

    def __rsub__(self, other):
        if isinstance(other, int):
            return other - self.rank

        if isinstance(other, Card):
            return other.rank - self.rank

        raise CardArithmeticError('cards can only subtract'
                                  'cards and integers')

    def __cmp__(self, other):
        if isinstance(other, Card):
            return cmp(self.rank, other.rank)

        if isinstance(other, int):
            return cmp(self.rank, other)

        raise CardComparisonError('cards can only be compared'
                                  'to cards and integers')

    def identical_to(self, other):
        if not isinstance(other, Card):
            raise CardComparisonError('cannot identity-check card'
                                      'with non-card')

        return (self.rank, self.suit) == (other.rank, other.suit)

    @classmethod
    def from_chars(cls, chars):
        """return a new object from a pair of character symbols"""

        try:
            if len(chars) != 2:
                raise CardCreationError('invalid card symbols')

            rank = chars[0].upper()
            suit = chars[1].lower()
            return cls(cls.ranks.index(rank), cls.suits.index(suit))

        except (TypeError, AttributeError, ValueError):
            raise CardCreationError('invalid card symbols')